require 'sketchup.rb'

#=============================================================================
# SU_GridDialog Class
#=============================================================================
class SU_GridDialog < UI::WebDialog

@@CheckBoxChecked = "checked"

def initialize()
  @@dialog = self
  @dialogKey = "SU_GridDialog"
  super("SU_Grid", false, @dialogKey, 400, 460, 100, 100, false) 

  width = SU_Grid.default_width()
  spacing = SU_Grid.default_spacing()
  offset = SU_Grid.default_offset()

  @width_string = (width.to_l).to_s
  @spacing_string = (spacing.to_l).to_s

  offsetval = offset.to_f
  if (offsetval < 0)
    offsetval = -offsetval
    @offset_string = "-" + (offsetval.to_l).to_s
  else
    @offset_string = (offsetval.to_l).to_s
  end

  @width_string.sub!(/"/, "&quot;") 
  @spacing_string.sub!(/"/, "&quot;") 
  @offset_string.sub!(/"/, "&quot;") 
  
  # Assign Callbacks
  add_action_callback("on_cancel") { |d, params| cancel() }
  add_action_callback("on_flip")  { |d, params| flip(params) }
  add_action_callback("on_delete_grid")  { |d, params| delete_grid(params) }
  add_action_callback("on_show_all") { |d, params| show_all() }
  add_action_callback("on_hide_all") { |d, params| hide_all() }
  add_action_callback("on_delete_all") { |d, params| delete_all() }
  add_action_callback("on_create_grid") { |d, params| create_grid(params) }

  # Generate html
  set_html(build_webpage)
end


def single_face_selected
  ss = Sketchup.active_model.selection
  return nil if (ss.empty?)
  faces = ss.find_all {|e| e.kind_of? Sketchup::Face}
  return nil  if (faces.length != 1)
  return faces[0]
end


def build_planelist
  # default is face if a single face is selected; else
  # default is the first type that is not already created
  #
  xy_checked = yz_checked = zx_checked = face_checked = ""
  if (single_face_selected())
    @planetype = 'Face'
    face_checked = @@CheckBoxChecked
  else 
    counts = Hash.new{0.to_i}
    @gridlist.each{|name, grid|
      gridtype = grid.get_attribute('SU_Grid', 'gridtype')
      counts[gridtype] += 1
    }
    if (counts['XY'] == 0)
      @planetype = 'XY'
      xy_checked = @@CheckBoxChecked
    elsif (counts['YZ'] == 0)
      @planetype = 'YZ'
      yz_checked = @@CheckBoxChecked
    elsif (counts['ZX'] == 0)
      @planetype = 'ZX'
      zx_checked = @@CheckBoxChecked
    else
      @planetype = 'XY'
      xy_checked = @@CheckBoxChecked
    end
  end

  html = <<-END_HERE
        <li> <input type="radio" name="plane_orientation" onclick="change_type();" value="XY" #{xy_checked}/> XY </li>
        <li> <input type="radio" name="plane_orientation" onclick="change_type();" value="YZ" #{yz_checked}/> YZ </li>
        <li> <input type="radio" name="plane_orientation" onclick="change_type();" value="ZX" #{zx_checked}/> ZX </li>
        <li> <input type="radio" name="plane_orientation" onclick="change_type();" value="Face" #{face_checked}/> Face </li>
END_HERE
  return html
end


def unique_name(base)
  name = base
  grid = @gridlist[name]
  count = 0.to_i
  while (grid)
    count += 1
    name = base + "_" + count.to_s
    grid = @gridlist[name]
  end
  return name
end


def build_planeargslist
  default_name = unique_name(@planetype)
  if (@planetype == :Face)
    html = <<-WIDTH_DISABLED
        <li>Width:<input id="width" type="text" style="width: 14ex;" maxlength="40" 
          style="background-color: rgb(236,233,216);" readonly="readonly" value='' /></li>
WIDTH_DISABLED
  else
    html = <<-WIDTH_ENABLED
        <li>Width:<input id="width" type="text" style="width: 14ex;" maxlength="40" 
          value="#{@width_string}" /></li>
WIDTH_ENABLED
  end
  html += <<-END_HERE
        <li>Spacing:<input id="spacing" type="text" value="#{@spacing_string}" style="width: 14ex;" maxlength="40" /></li>
        <li>Offset:<input id="offset" type="text" value="#{@offset_string}" style="width: 14ex;" maxlength="40" /></li>
        <li>Name:<input id="name" type="text" value="#{default_name}" style="width: 14ex;" maxlength="40" /></li>
END_HERE
  return html
end


def build_gridlist
  # Make an html entry for each
  html = ""
  @gridlist.each{|name, grid|
    entry = <<-END_HERE
    <tr>
    <td width="40%">
      #{name}
    </td>
    <td width="30%" style="text-align:center;">
      <input type="checkbox" onclick="flip('#{name}');" #{ 'checked' if ( grid.visible? ) } />
    </td>
    <td width="30%" style="text-align:center;">
      <button style="font-size:50%; font-weight:bold" onclick="delete_grid('#{name}');">X</button>
    </td>
    </tr>
END_HERE
    html += entry
  }
  return html
end


def build_webpage
  # Find grids in model
  @gridlist = SU_Grid.find_grids()

  # Build html list of plane choices
  planelist_html = self.build_planelist
  
  # Build html list of plane arguments
  planeargs_html = self.build_planeargslist
  
  # Build html list of existing grids 
  gridlist_html = self.build_gridlist

  # Build the web page
html = <<-END_HERE
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
  <meta http-equiv="Content-Type" content="text/html" />
  <title>SU Animate</title>

  <style type="text/css">
    * {padding: 0px; margin: 0px;}
    html {overflow: noscroll;}
    body {font: 80% Arial, sans-serif; padding: 10px; background-color: rgb(236,233,216);}
    ul {list-style: none; font-weight: normal;}
  </style>

  <script type="text/javascript">
    function cancel() {window.location = 'skp:on_cancel';}
    function show_all() {window.location = 'skp:on_show_all';}
    function hide_all() {window.location = 'skp:on_hide_all';}
    function delete_all() {window.location = 'skp:on_delete_all';}

    function flip(aname) {
      window.location = 'skp:on_flip@' + aname;
    }

    function delete_grid(aname) {
      window.location = 'skp:on_delete_grid@' + aname;
    }

    function change_type() {
      var gridtype = "XY";
      var elems = document.forms[0].elements;
      for (var i = 0; i < elems.length; i++) {
        var elem = elems[i];
        if ( elem.type == "radio" && elem.checked) {
          gridtype = elem.value;
        }
      }
      document.getElementById('name').value = gridtype;
      elem = document.getElementById('width')
      if (gridtype == "Face") {
        elem.readOnly = true;
        elem.style.backgroundColor = 'rgb(236,233,216)';
        if (elem.value != "") {
          elem.defaultValue = elem.value;
          elem.value = "";
        }
      } else {
        elem.readOnly = false;
        elem.style.backgroundColor = 'white';
        if (elem.value == "") {
          if (elem.defaultValue == "") {
            elem.value = "#{@width_string}";
          } else {
            elem.value = elem.defaultValue;
          }
        }
      }
      
    }

    function create_grid() {
      var gridtype = "XY";
      var elems = document.forms[0].elements;
      for (var i = 0; i < elems.length; i++) {
        var elem = elems[i];
        if ( elem.type == "radio" && elem.checked) {
          gridtype = elem.value;
        }
      }

      var name = document.getElementById('name').value;
      
      // Replace the "'" (feet) token, screws up the string stuff 
      var width = document.getElementById('width').value.replace(/'/,":");
      var spacing = document.getElementById('spacing').value.replace(/'/,":");
      var offset = document.getElementById('offset').value.replace(/'/,":");
      window.location = 'skp:on_create_grid@' + gridtype + "," + name + "," + width + "," + spacing + "," + offset;
    }
    
    document['onkeydown'] = keyPressHandler
    function keyPressHandler(e) {
      e = e || event 
      var kC  = (window.event) ?    // MSIE or Firefox?
                 event.keyCode : e.keyCode;
      var Esc = (window.event) ?   
                27 : e.DOM_VK_ESCAPE // MSIE : Firefox
      if(kC==Esc) {cancel();}
    }
    
  </script>
  </head>

  <body> 
  <fieldset style="padding:0em 1em 1em 1em">
  <form>
  <legend>Define Grids </legend>
  <table border="0" width="100%" cellpadding="0">
    <tr>
    <td width="40%" valign="top">
      <ul> 
        #{planelist_html}
      </ul>
    </td>
    <td width="60%" valign="top" style="text-align:right;">
      <ul> 
        #{planeargs_html}
      </ul>
    </td>
    </tr>
  </table>
  </form>
  <table border="0" width="100%" cellpadding="0">
    <tr>
    <td width="100%" valign="top" style="text-align:center; padding:0.5em 0em 0em 0em">
      <button onclick="create_grid();">Create Grid</button>
    </td>
    </tr>
  </table>
  </fieldset>
  <fieldset style="padding:1em 1em 1em 1em;">
  <legend>Show/Hide Grids</legend>
  <table id="gridlist" border="0" width="100%" cellpadding="0" style="font-size:80%;"; >
    <tr>
    <td width="40%">
      Name
    </td>
    <td width="30%" style="text-align:center;">
      Visible
    </td>
    <td width="30%" style="text-align:center;">
      Delete
    </td>
    </tr>
  </table>
  <table id="gridlist" border="0" width="100%" cellpadding="0" bgcolor="white">
    #{gridlist_html}
  </table>
  <table border="0" width="100%" cellpadding="0">
    <tr>
    <td width="100%" valign="top" style="text-align:center; padding:0.5em 0em 0em 0em">
      <button onclick="show_all();">Show All</button>
      <button onclick="hide_all();">Hide All</button>
      <button onclick="delete_all();">Delete All</button>
    </td>
    </tr>
  </table>
  </fieldset>
  <div align="center" style="color:#3399FF;margin:1em 0em 0em 0em">
    <a href="http://www.suplugins.com/" target="_blank" >www.suplugins.com</a> 
  </div>
</html>
END_HERE
  return html
end


def refresh_dialog()
  close
  @@dlg = SU_GridDialog.new()
  @@dlg.show 
end

def flip(name)
  grid = @gridlist[name]
  return if not grid
  return if grid.deleted?
  grid.visible = ! grid.visible?
end

def delete_grid(name)
  grid = @gridlist[name]
  return if not grid
  return if grid.deleted?
  grid.erase!
  refresh_dialog() 
end

def create_grid(params)
  gridtype, name, width_str, spacing_str, offset_str = params.split(',')

  # Restore the "'" (feet) token
  width_str.sub!(/:/, "'") 
  spacing_str.sub!(/:/, "'") 
  offset_str.sub!(/:/, "'") 

  width = width_str.to_l.to_f
  spacing = spacing_str.to_l.to_f
  offset = offset_str.to_l.to_f
 
  # Check name for uniqueness
  grid = @gridlist[name]
  if (grid)
    UI.messagebox("There is already a grid named \"#{name}\". \n Please select a unique name", MB_OK)
    return
  end

  # To make a grid on a face, a single face must be selected
  if (gridtype == "Face" && (not single_face_selected()))
    UI.messagebox("Select a face", MB_OK)
    return
  end
  
  SU_Grid.create_grid(gridtype, width, spacing, offset, name)

  refresh_dialog() 
end

def cancel()
  @@dialog = nil
  close
end

def show_all()
  SU_Grid.show_all
  refresh_dialog() 
end

def hide_all()
  SU_Grid.hide_all
  refresh_dialog() 
end

def delete_all()
  SU_Grid.delete_all
  refresh_dialog() 
end

def SU_GridDialog.refresh_dialog()
  return nil if not @@dialog
  @@dialog.refresh_dialog() 
end


end # class SU_GridDialog


#=============================================================================
# SU_GridObserver Class
#=============================================================================
class SU_GridObserver < Sketchup::EntityObserver
def onEraseEntity(entity)
  SU_GridDialog.refresh_dialog()
end


end # class SU_GridObserver 

     
#=============================================================================
# SU_Grid Class
#=============================================================================
class SU_Grid

# FIX THIS allow user to set these
# FIX THIS make it depend on units
@@width = 360
@@spacing = 12
@@offset = 0

# Find grids in model
def SU_Grid.find_grids()
  grids = Hash.new
  # iterate through the model
  for ent in Sketchup.active_model.active_entities
    next if not ent.is_a? Sketchup::Group
    name = ent.get_attribute('SU_Grid', 'name')
    next if not name
    grids[name] = ent
  end
  return grids
end

def SU_Grid.delete_all()
  eraseme = []
  for ent in Sketchup.active_model.active_entities
    next if not ent.is_a? Sketchup::Group
    name = ent.get_attribute('SU_Grid', 'name')
    next if not name
    eraseme.push ent
  end
  
  eraseme.each {|ent|
    ent.erase!
  }
end

def SU_Grid.set_all_visible(value)
  for ent in Sketchup.active_model.active_entities
    next if not ent.is_a? Sketchup::Group
    name = ent.get_attribute('SU_Grid', 'name')
    next if not name
    ent.visible = value
  end
end

def SU_Grid.show_all() SU_Grid.set_all_visible(true) end
def SU_Grid.hide_all() SU_Grid.set_all_visible(false) end
def SU_Grid.default_width() return @@width end
def SU_Grid.default_spacing() return @@spacing end
def SU_Grid.default_offset() return @@offset end

def SU_Grid.create_grid(gridtype, width, spacing, offset, name)
  @@width = width
  @@spacing = spacing
  @@offset = offset
  if (gridtype == "Face")
    SU_Grid.create_face_grid(spacing, offset, name)
  else
    SU_Grid.create_plane_grid(gridtype, width, spacing, offset, name)
  end
end


def SU_Grid.create_face_grid(spacing, offset, name)
  # get transform that describes selected face
  ss = Sketchup.active_model.selection
  norm = ss[0].normal
  origin = ss[0].edges[0].line[0]
  xaxis = ss[0].edges[0].line[1] # Vector3D of first edge becomes x axis
  yaxis = xaxis * norm

  if (offset != 0)
    offset_vec = norm
    if (offset < 0)
      offset_vec.length = -offset
    else
      offset_vec.length = offset
    end
    origin = origin + offset_vec
  end
  
  xform = Geom::Transformation.new(xaxis, yaxis, norm, origin)
  inv_xform = xform.inverse

  vertices = ss[0].vertices
  normalized_vertices = vertices.collect { |x| inv_xform * x.position }

  x_coords = normalized_vertices.collect { |v| v.x }
  y_coords = normalized_vertices.collect { |v| v.y }
  x_coords.sort!
  y_coords.sort!
  sorted_vertices = vertices.sort {|x,y| x.position.distance(origin) <=> y.position.distance(origin) }
  last = normalized_vertices.length - 1

  SU_Grid.draw_grid('Face', name,
    x_coords[0], x_coords[last],
    y_coords[0], y_coords[last],
    spacing.to_f, xform)
end

def SU_Grid.create_plane_grid(gridtype, width, spacing, offset, name)
  if (gridtype == 'XY')
    xform = Geom::Transformation.new(Geom::Point3d.new(0,0,offset), Geom::Vector3d.new(1,0,0),  Geom::Vector3d.new(0,1,0))
  elsif (gridtype == 'YZ')
    xform = Geom::Transformation.new(Geom::Point3d.new(offset,0,0), Geom::Vector3d.new(0,1,0),  Geom::Vector3d.new(0,0,1))
  else
    xform = Geom::Transformation.new(Geom::Point3d.new(0,offset,0), Geom::Vector3d.new(0,0,1),  Geom::Vector3d.new(1,0,0))
  end

  length = width
  SU_Grid.draw_grid(gridtype, name,
    -width.to_f/2, width.to_f/2,
    -length.to_f/2, length.to_f/2,
    spacing.to_f, 
    xform)
end

def SU_Grid.draw_grid(gridtype, name,
  start_x, stop_x, start_y, stop_y, interval, xform)
  model = Sketchup.active_model
  model.start_operation "Create Grid"

  group = model.active_entities.add_group
  entities = group.entities

  step_x = start_x
  while step_x <= stop_x
    start_pt = [step_x, start_y, 0]
    stop_pt = [step_x, stop_y, 0]
    line1 = entities.add_cline(xform*start_pt, xform*stop_pt, "." )
    step_x = step_x + interval
  end
  start_pt = [stop_x, start_y, 0]
  stop_pt = [stop_x, stop_y, 0]
  line1 = entities.add_cline(xform*start_pt, xform*stop_pt, "." )

  step_y = start_y
  while step_y <= stop_y
    start_pt = [start_x, step_y, 0]
    stop_pt = [stop_x, step_y, 0]
    line1 = entities.add_cline(xform*start_pt, xform*stop_pt, "." )
    step_y = step_y + interval
  end
  start_pt = [start_x, stop_y, 0]
  stop_pt = [stop_x, stop_y, 0]
  line1 = entities.add_cline(xform*start_pt, xform*stop_pt, "." )

  group.set_attribute('SU_Grid', 'gridtype', gridtype)
  group.set_attribute('SU_Grid', 'name', name)
  layer = Sketchup.active_model.layers.add "SU_Grid"
  group.layer = layer
  group.name = name
  
  # FIX THIS: create grid, esc, new, create grid, esc, delete grid, new, close -> SPLAT!
  # group.add_observer(SU_GridObserver.new)
  model.commit_operation
  model.active_view.invalidate
end

end # class SU_Grid


def grid_dialog()
  dlg = SU_GridDialog.new()
  dlg.show 
end


def gd
  grid_dialog()
end


# -----------------------------------------------------------------------------

if( not file_loaded?("su_grid.rb") )

  # Menu
  add_separator_to_menu("Draw")
  UI.menu("Draw").add_item("Construction Grid") { grid_dialog }

  # Context Menu
  UI.add_context_menu_handler do |menu|
    menu.add_separator
    menu.add_item("Construction Grid") { grid_dialog }
  end
end

# -----------------------------------------------------------------------------
file_loaded("su_grid.rb")
